package org.bitbucket.bradleysmithllc.etlagent.informatica.provider;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import org.bitbucket.bradleysmithllc.etlagent.informatica.dto.RemoteFile;
import org.bitbucket.bradleysmithllc.etlagent.informatica.dto.json.DomainsProperty;
import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.Incomplete;
import org.bitbucket.bradleysmithllc.etlunit.util.VelocityUtil;
import org.bitbucket.bradleysmithllc.json.validator.JsonUtils;

import java.io.File;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

public class InformaticaDomain {
	private final String domainName;
	private final String clientVersion;
	private final String connectivityHost;
	private final int connectivityPort;
	private final String userName;
	private final String passwordEncrypted;
	private final String securityDomain;

	private final File domainFile;

	private final File informaticaBinDirectory;
	private final File workingRoot;

	private final RuntimeSupport runtimeSupport;

	private final InformaticaRepository defaultRepository;
	private final InformaticaWebServicesHub defaultWebServicesHub;
	private final DomainsProperty.ClientType clientType;

	private int numRequestsSent = 0;

	private final InformaticaIntegrationService defaultIntegrationService;

	private final List<InformaticaRepository> repositories = new ArrayList<InformaticaRepository>();

	private final String agentHost;
	private final int agentPort;

	@Incomplete(explanation = "Implement domains.infa as a user-supplied file")
	public InformaticaDomain(String name, JsonNode value, RuntimeSupport runtimeSupport, File domainsFile)
	{
		domainName = name;
		clientVersion = getRequiredAttribute("client-version", value);
		connectivityHost = getOptionalAttribute("connectivity-host", value);
		securityDomain = getOptionalAttribute("security-domain", value);

		this.runtimeSupport = runtimeSupport;

		String cltype = getOptionalAttribute("client-type", value);

		clientType = resolveType(cltype);

		switch (clientType) {
			case AGENT:
				// agent is a required property now
				ObjectNode agentNode = (ObjectNode) value.get("agent");

				if (agentNode == null)
				{
					throw new IllegalArgumentException("agent property is required when client type is 'agent'");
				}

				JsonNode hostNameNode = agentNode.get("host-name");

				if (hostNameNode == null)
				{
					throw new IllegalArgumentException("agent object requires host-name");
				}

				agentHost = hostNameNode.asText();

				JsonNode jsonNode = agentNode.get("host-port");

				if (jsonNode != null)
				{
					agentPort = jsonNode.asInt();
				}
				else
				{
					agentPort = 8188;
				}
				break;
			case MOCK:
			case LOCAL:
			default:
				agentHost = null;
				agentPort = 0;
				break;
		}

		String root = getRequiredAttribute("working-root", value);

		workingRoot = new File(root);

		String binRoot = getOptionalAttribute("informatica-bin-directory", value);

		if (binRoot != null)
		{
			informaticaBinDirectory= new File(binRoot);
		}
		else
		{
			informaticaBinDirectory = null;
		}

		String port = getOptionalAttribute("connectivity-port", value);

		if (port != null)
		{
			connectivityPort = Integer.parseInt(port);
		}
		else
		{
			connectivityPort = -1;
		}

		userName = getRequiredAttribute("username", value);
		passwordEncrypted = getRequiredAttribute("password-encrypted", value);

		JsonNode erepositories = JsonUtils.query(value, "repositories");

		if (erepositories == null)
		{
			throw new IllegalArgumentException("Informatica domain missing repositories");
		}

		if (!erepositories.isObject())
		{
			throw new IllegalArgumentException("Informatica repositories not an object: " + name);
		}

		Iterator<Map.Entry<String,JsonNode>> it = erepositories.fields();

		while (it.hasNext())
		{
			Map.Entry<String, JsonNode> repoEntry = it.next();
			InformaticaRepository repository = new InformaticaRepository(this, repoEntry.getKey(), runtimeSupport);

			if (getRepository(repoEntry.getKey()) != null)
			{
				throw new IllegalArgumentException("Duplicate repository declaration: " + repoEntry.getKey());
			}

			repositories.add(repository);

			// load integration services
			JsonNode intSvcs = JsonUtils.query(repoEntry.getValue(), "integration-services");

			if (intSvcs != null)
			{
				if (intSvcs.isArray())
				{
					Iterator<JsonNode> iit = intSvcs.iterator();

					while (iit.hasNext())
					{
						String intSvcName = iit.next().asText();

						if (getIntegrationService(intSvcName) != null)
						{
							throw new IllegalArgumentException("Integration service " + intSvcName + " specified more than once");
						}

						repository.getIntegrationServices().add(new InformaticaIntegrationService(intSvcName, repository, runtimeSupport));
					}
				}
				else
				{
					throw new IllegalArgumentException("Integration services property must be an array");
				}
			}

			// load integration services
			JsonNode webSvcs = JsonUtils.query(repoEntry.getValue(), "web-services-hubs");

			if (webSvcs != null)
			{
				if (webSvcs.isObject())
				{
					Iterator<Map.Entry<String,JsonNode>> iit = webSvcs.fields();

					while (iit.hasNext())
					{
						Map.Entry<String, JsonNode> webSvcHub = iit.next();

						if (getWebServicesHub(webSvcHub.getKey()) != null)
						{
							throw new IllegalArgumentException("Web services hub " + webSvcHub + " specified more than once");
						}

						JsonNode webSvcHubValue = webSvcHub.getValue();
						String hostName = getOptionalAttribute("host-name", webSvcHubValue);

						if (hostName == null)
						{
							try {
								hostName = InetAddress.getLocalHost().getHostAddress();
							} catch (UnknownHostException e) {
								throw new RuntimeException("Could not determine local host name", e);
							}
						}

						String hostPort = getOptionalAttribute("host-port", webSvcHubValue);

						if (hostPort == null)
						{
							hostPort = "7333";
						}

						repository.getWebServicesHubs().add(new InformaticaWebServicesHub(webSvcHub.getKey(), repository, hostName, Integer.parseInt(hostPort), runtimeSupport));
					}
				}
				else
				{
					throw new IllegalArgumentException("Web services hub property must be an object");
				}
			}
		}

		if (repositories.size() == 0)
		{
			throw new IllegalArgumentException("Informatica domain repositories element is empty");
		}

		String defRepo = getOptionalAttribute("default-repository", value);

		InformaticaRepository repo = repositories.get(0);

		if (defRepo != null)
		{
			repo = getRepository(defRepo);

			if (repo == null)
			{
				throw new IllegalArgumentException("Default repository does not exist in domain: " + defRepo);
			}
		}

		defaultRepository = repo;

		String defIs = getOptionalAttribute("default-integration-service", value);

		InformaticaIntegrationService dis = getDefaultRepository().getDefaultIntegrationService();

		if (defIs != null)
		{
			dis = getIntegrationService(defIs);

			if (dis == null)
			{
				throw new IllegalArgumentException("Default integration service not found in domain: " + defIs);
			}
		}

		defaultIntegrationService = dis;

		String defIsh = getOptionalAttribute("default-web-services-hub", value);

		InformaticaWebServicesHub dish = getDefaultRepository().getDefaultWebServicesHub();

		if (defIsh != null)
		{
			dish = getWebServicesHub(defIsh);

			if (dish == null)
			{
				throw new IllegalArgumentException("Default web services hub not found in domain: " + defIsh);
			}
		}

		defaultWebServicesHub = dish;

		if (getConnectivityHost() != null) {
			String template = null;
			try
			{
				template = IOUtils.readURLToString(getClass().getResource("/domains.infa.vm"));

				Map<String, String> bean = new HashMap<String, String>();
				bean.put("domainName", getDomainName());
				bean.put("connectivityHost", getConnectivityHost());
				bean.put("connectivityPort", String.valueOf(getConnectivityPort()));

				String vel = VelocityUtil.writeTemplate(template, bean);

				domainFile = runtimeSupport.createTempFile("domains.infa." + getDomainName());

				IOUtils.writeBufferToFile(domainFile, new StringBuffer(vel));
			}
			catch (Exception e)
			{
				throw new IllegalStateException(e);
			}
		}
		else
		{
			domainFile = domainsFile;
		}
	}

	private DomainsProperty.ClientType resolveType(String name) {
		if (name == null)
		{
			return DomainsProperty.ClientType.LOCAL;
		}

		return DomainsProperty.ClientType.fromValue(name);
	}

	private InformaticaWebServicesHub getWebServicesHub(String name) {
		for (InformaticaRepository irepo : repositories)
		{
			for (InformaticaWebServicesHub iis : irepo.getWebServicesHubs())
			{
				if (iis.getName().equals(name))
				{
					return iis;
				}
			}
		}

		return null;
	}

	public String getSecurityDomain() {
		return securityDomain;
	}

	public File getInformaticaBinDirectory() {
		return informaticaBinDirectory;
	}

	private String getOptionalAttribute(String s, JsonNode value) {
		return getAttribute(s, value, false);
	}

	private String getRequiredAttribute(String s, JsonNode value) {
		return getAttribute(s, value, true);
	}

	private String getAttribute(String s, JsonNode value, boolean required) {
		JsonNode res = JsonUtils.query(value, s);

		if (required && res == null)
		{
			throw new IllegalArgumentException("Informatica domain missing attribute: " + s);
		}

		return res == null ? null : res.asText();
	}

	public String getDomainName() {
		return domainName;
	}

	public String getClientVersion() {
		return clientVersion;
	}

	public String getConnectivityHost() {
		return connectivityHost;
	}

	public int getConnectivityPort() {
		return connectivityPort;
	}

	public String getUserName() {
		return userName;
	}

	public String getPasswordEncrypted() {
		return passwordEncrypted;
	}

	public InformaticaRepository getRepository(String name)
	{
		for (InformaticaRepository repo : repositories)
		{
			if (repo.getRepositoryName().equals(name))
			{
				return repo;
			}
		}

		return null;
	}

	public List<InformaticaRepository> getRepositories() {
		return repositories;
	}

	public InformaticaRepository getDefaultRepository() {
		return defaultRepository;
	}

	public File getWorkingRoot() {
		return workingRoot;
	}

	public File getLocalizedWorkingRoot()
	{
		return new FileBuilder(workingRoot).subdir(runtimeSupport.getProjectUID()).file();
	}

	public InformaticaIntegrationService getDefaultIntegrationService()
	{
		return defaultIntegrationService;
	}

	public InformaticaWebServicesHub getDefaultWebServicesHub() {
		return defaultWebServicesHub;
	}

	public InformaticaIntegrationService getIntegrationService(String isName)
	{
		for (InformaticaRepository irepo : repositories)
		{
			for (InformaticaIntegrationService iis : irepo.getIntegrationServices())
			{
				if (iis.getIntegrationServiceName().equals(isName))
				{
					return iis;
				}
			}
		}

		return null;
	}

	public boolean containsIntegrationService(String isName)
	{
		return getIntegrationService(isName) != null;
	}

	public File getDomainFile()
	{
		if (domainFile == null)
		{
			throw new UnsupportedOperationException("Domain file not supported yet");
		}

		return domainFile;
	}

	public DomainsProperty.ClientType getClientType() {
		return clientType;
	}

	public String getAgentHost() {
		return agentHost;
	}

	public int getAgentPort() {
		return agentPort;
	}

	public void dispose()
	{
		for (InformaticaRepository repository : getRepositories())
		{
			repository.dispose();
		}
	}

	File getInformaticaBinDir()
	{
		// get the client version for the binary
		String ver = getClientVersion();
		// construct a path to the power center binaries using a well-known path combined with the version number

		// check for an explicit bin dir, then use the etl unit dir
		if (informaticaBinDirectory != null)
		{
			return informaticaBinDirectory;
		}
		else
		{
			return new FileBuilder(runtimeSupport.getVendorBinaryDir()).subdir("PowerCenter").subdir(ver).subdir("server").subdir("bin").file();
		}
	}

	public File getWorkspaceDir(RemoteFile.fileType type)
	{
		return new FileBuilder(getLocalizedWorkingRoot()).mkdirs().subdir(type.getDirectoryName()).mkdirs().file();
	}

	public File getTargetFilesDir()
	{
		return getWorkspaceDir(RemoteFile.fileType.targetFiles);
	}

	public File getBadFilesDir()
	{
		return getWorkspaceDir(RemoteFile.fileType.badFiles);
	}

	public File getSessionLogDir()
	{
		return getWorkspaceDir(RemoteFile.fileType.sessionLog);
	}

	public File getWorkflowLogDir()
	{
		return getWorkspaceDir(RemoteFile.fileType.workflowLog);
	}

	public File getParameterFilesDir()
	{
		return getWorkspaceDir(RemoteFile.fileType.parameterFiles);
	}

	public File getSourceFilesDir()
	{
		return getWorkspaceDir(RemoteFile.fileType.sourceFiles);
	}

	public File getLookupFilesDir()
	{
		return getWorkspaceDir(RemoteFile.fileType.lookupFiles);
	}

	void request()
	{
		numRequestsSent++;
	}

	public int getNumRequestsSent() {
		return numRequestsSent;
	}

	@Override
	public String toString() {
		return "InformaticaDomain{" +
				"domainName='" + domainName + '\'' +
				'}';
	}
}